"""``pytest`` fixtures."""

from collections.abc import Iterator

import pytest
from mock_vws import MockVWS
from mock_vws.database import VuforiaDatabase
from vws import VWS, CloudRecoService


@pytest.fixture(name="mock_database")
def fixture_mock_database() -> Iterator[VuforiaDatabase]:
    """Yield a mock ``VuforiaDatabase``."""
    with MockVWS() as mock:
        database = VuforiaDatabase()
        mock.add_database(database=database)
        yield database


@pytest.fixture()
def vws_client(mock_database: VuforiaDatabase) -> VWS:
    """Return a VWS client which connects to a mock database."""
    return VWS(
        server_access_key=mock_database.server_access_key,
        server_secret_key=mock_database.server_secret_key,
    )


@pytest.fixture()
def cloud_reco_client(
    mock_database: VuforiaDatabase,
) -> CloudRecoService:
    """Return a ``CloudRecoService`` client which connects to a mock database."""
    return CloudRecoService(
        client_access_key=mock_database.client_access_key,
        client_secret_key=mock_database.client_secret_key,
    )



"""Tests for the VWS CLI help."""

from __future__ import annotations

import os
from pathlib import Path

import pytest
from click.testing import CliRunner
from vws_cli import vws_group
from vws_cli.query import vuforia_cloud_reco

_SUBCOMMANDS = [[item] for item in vws_group.commands]
_BASE_COMMAND: list[list[str]] = [[]]
_COMMANDS = _BASE_COMMAND + _SUBCOMMANDS


@pytest.mark.parametrize(
    "command",
    _COMMANDS,
    ids=[str(cmd) for cmd in _COMMANDS],
)
def test_vws_command_help(command: list[str]) -> None:
    """Expected help text is shown for ``vws`` commands.

    This help text is defined in files.
    To update these files, run the command ``bash admin/update_cli_tests.sh``.
    """
    runner = CliRunner()
    arguments = [*command, "--help"]
    result = runner.invoke(vws_group, arguments, catch_exceptions=False)
    assert result.exit_code == 0
    help_output_filename = "-".join(["vws", *command]) + ".txt"
    help_outputs_dir = Path(__file__).parent / "help_outputs"
    expected_help_file = help_outputs_dir / help_output_filename
    try:
        expected_help = expected_help_file.read_text()
        assert result.output == expected_help
    except (AssertionError, FileNotFoundError):  # pragma: no cover
        if os.getenv("FIX_CLI_TESTS") == "1":
            help_outputs_dir.mkdir(exist_ok=True)
            expected_help_file.touch()
            expected_help_file.write_text(result.output)
        else:
            raise


def test_query_help() -> None:
    """Expected help text is shown for the ``vuforia-cloud-reco`` command.

    This help text is defined in files.
    To update these files, run the command ``bash admin/update_cli_tests.sh``.
    """
    runner = CliRunner()
    arguments = ["--help"]
    result = runner.invoke(
        vuforia_cloud_reco,
        arguments,
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    help_output_filename = "vuforia_cloud_reco.txt"
    help_outputs_dir = Path(__file__).parent / "help_outputs"
    expected_help_file = help_outputs_dir / help_output_filename
    try:
        expected_help = expected_help_file.read_text()
        assert result.output == expected_help
    except (AssertionError, FileNotFoundError):  # pragma: no cover
        if os.getenv("FIX_CLI_TESTS") == "1":
            help_outputs_dir.mkdir(exist_ok=True)
            expected_help_file.touch()
            expected_help_file.write_text(result.output)
        else:
            raise



"""Release the next version."""

from pathlib import Path

from admin.binaries import make_linux_binaries


def main() -> None:
    """Perform a release."""
    make_linux_binaries(repo_root=Path("."))


if __name__ == "__main__":
    main()



"""``click`` options regarding targets."""

from __future__ import annotations

import functools
from collections.abc import Callable
from enum import Enum
from pathlib import Path

import click


def target_id_option(command: Callable[..., None]) -> Callable[..., None]:
    """An option decorator for choosing a target ID."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--target-id",
        type=str,
        help="The ID of a target in the Vuforia database.",
        required=True,
    )
    function: Callable[..., None] = click_option_function(command)
    return function


def target_name_option(
    command: Callable[..., None] | None,
    *,
    required: bool,
) -> Callable[..., None]:
    """An option decorator for choosing a target name."""
    if not command:
        # Ignore type error as per https://github.com/python/mypy/issues/1484.
        return functools.partial(  # type: ignore[return-value]
            target_name_option,
            required=required,
        )
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--name",
        type=str,
        help="The name of the target in the Vuforia database.",
        required=required,
    )
    function: Callable[..., None] = click_option_function(command)
    return function


def target_width_option(
    command: Callable[..., None] | None,
    *,
    required: bool,
) -> Callable[..., None]:
    """An option decorator for choosing a target width."""
    if not command:
        # Ignore type error as per https://github.com/python/mypy/issues/1484.
        return functools.partial(  # type: ignore[return-value]
            target_width_option,
            required=required,
        )
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--width",
        type=float,
        help="The width of the target in the Vuforia database.",
        required=required,
    )
    assert command is not None
    function: Callable[..., None] = click_option_function(command)
    return function


def target_image_option(
    command: Callable[..., None] | None,
    *,
    required: bool,
) -> Callable[..., None]:
    """An option decorator for choosing a target image."""
    if not command:
        # Ignore type error as per https://github.com/python/mypy/issues/1484.
        return functools.partial(  # type: ignore[return-value]
            target_image_option,
            required=required,
        )

    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--image",
        "image_file_path",
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=False,
            path_type=Path,
        ),
        help="The path to an image to upload and set as the target image.",
        required=required,
    )

    function: Callable[..., None] = click_option_function(command)
    return function


class ActiveFlagChoice(Enum):
    """Choices for active flag."""

    TRUE = "true"
    FALSE = "false"


def _active_flag_choice_callback(
    ctx: click.core.Context,
    param: click.core.Option | click.core.Parameter,
    value: str | None,
) -> ActiveFlagChoice | None:
    """Use as a callback for active flag options."""
    # This is to satisfy the "vulture" linter.
    assert ctx
    assert param

    if value is None:
        return None

    return ActiveFlagChoice(value)


def active_flag_option(
    command: Callable[..., None] | None,
    *,
    allow_none: bool,
) -> Callable[..., None]:
    """An option decorator for setting a target's active flag."""
    if not command:
        # Ignore type error as per https://github.com/python/mypy/issues/1484.
        return functools.partial(  # type: ignore[return-value]
            active_flag_option,
            allow_none=allow_none,
        )

    if allow_none:
        default = None
        show_default = False
    else:
        default = ActiveFlagChoice.TRUE.value
        show_default = True

    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--active-flag",
        "active_flag_choice",
        help="Whether or not the target is active for query.",
        type=click.Choice([item.value for item in ActiveFlagChoice]),
        default=default,
        callback=_active_flag_choice_callback,
        show_default=show_default,
    )
    function: Callable[..., None] = click_option_function(command)
    return function


def application_metadata_option(
    command: Callable[..., None] | None = None,
) -> Callable[..., None]:
    """An option decorator for setting application metadata."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--application-metadata",
        type=str,
        required=False,
        help=(
            "The base64 encoded application metadata associated with the "
            "target."
        ),
    )
    assert command is not None
    function: Callable[..., None] = click_option_function(command)
    return function



"""A CLI for Vuforia Web Services."""

from pathlib import Path

import click
from setuptools_scm import get_version

from vws_cli.commands import (
    add_target,
    delete_target,
    get_database_summary_report,
    get_duplicate_targets,
    get_target_record,
    get_target_summary_report,
    list_targets,
    update_target,
    wait_for_target_processed,
)

_CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}

try:
    __version__ = get_version(root="../..", relative_to=__file__)
except LookupError:  # pragma: no cover
    # When pkg_resources and git tags are not available,
    # for example in a PyInstaller binary,
    # we write the file ``_setuptools_scm_version.py`` on ``pip install``.
    _VERSION_FILE = Path(__file__).parent / "_setuptools_scm_version.txt"
    __version__ = _VERSION_FILE.read_text()


@click.group(name="vws", context_settings=_CONTEXT_SETTINGS)
# We set the ``version`` parameter because in PyInstaller binaries,
# ``pkg_resources`` is not available.
#
# Click uses ``pkg_resources`` to determine the version if it is not given.
@click.version_option(version=__version__)
def vws_group() -> None:
    """Manage a Vuforia Web Services cloud database."""


vws_group.add_command(add_target)
vws_group.add_command(delete_target)
vws_group.add_command(get_database_summary_report)
vws_group.add_command(get_duplicate_targets)
vws_group.add_command(get_target_record)
vws_group.add_command(get_target_summary_report)
vws_group.add_command(list_targets)
vws_group.add_command(update_target)
vws_group.add_command(wait_for_target_processed)



"""Test for the Cloud Reco Service commands."""

from __future__ import annotations

import io
import uuid
from pathlib import Path
from textwrap import dedent

import yaml
from click.testing import CliRunner
from mock_vws import MockVWS
from mock_vws.database import VuforiaDatabase
from vws import VWS
from vws_cli.query import vuforia_cloud_reco


class TestQuery:
    """Tests for making image queries."""

    @staticmethod
    def test_no_matches(
        mock_database: VuforiaDatabase,
        tmp_path: Path,
        high_quality_image: io.BytesIO,
    ) -> None:
        """An empty list is returned if there are no matches."""
        runner = CliRunner(mix_stderr=False)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands: list[str] = [
            str(new_file),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        result_data = yaml.safe_load(result.stdout)
        assert result_data == []

    @staticmethod
    def test_matches(
        tmp_path: Path,
        high_quality_image: io.BytesIO,
        vws_client: VWS,
        mock_database: VuforiaDatabase,
    ) -> None:
        """Details of matching targets are shown."""
        name = uuid.uuid4().hex
        target_id = vws_client.add_target(
            name=name,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)

        runner = CliRunner(mix_stderr=False)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        result_data = yaml.safe_load(result.stdout)
        [matching_target] = result_data
        target_timestamp = matching_target["target_data"]["target_timestamp"]
        expected_result_data = {
            "target_data": {
                "application_metadata": None,
                "name": name,
                "target_timestamp": target_timestamp,
            },
            "target_id": target_id,
        }
        assert matching_target == expected_result_data

    @staticmethod
    def test_image_file_is_dir(
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """An appropriate error is given if the given image file path points to a
        directory.
        """
        runner = CliRunner(mix_stderr=False)
        commands: list[str] = [
            str(tmp_path),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        assert result.stdout == ""
        expected_stderr = dedent(
            f"""\
            Usage: vuforia-cloud-reco [OPTIONS] IMAGE
            Try 'vuforia-cloud-reco --help' for help.

            Error: Invalid value for 'IMAGE': File '{tmp_path}' is a directory.
            """,
        )
        assert result.stderr == expected_stderr

    @staticmethod
    def test_relative_path(
        tmp_path: Path,
        mock_database: VuforiaDatabase,
        high_quality_image: io.BytesIO,
    ) -> None:
        """Image file paths are resolved."""
        runner = CliRunner(mix_stderr=False)
        new_filename = uuid.uuid4().hex
        original_image_file = tmp_path / "foo"
        image_data = high_quality_image.getvalue()
        original_image_file.write_bytes(image_data)
        commands: list[str] = [
            str(new_filename),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        with runner.isolated_filesystem():
            new_file = Path(new_filename)
            new_file.symlink_to(original_image_file)
            result = runner.invoke(
                vuforia_cloud_reco,
                commands,
                catch_exceptions=False,
            )

        assert result.exit_code == 0
        result_data = yaml.safe_load(result.stdout)
        assert result_data == []

    @staticmethod
    def test_image_file_does_not_exist(
        mock_database: VuforiaDatabase,
        tmp_path: Path,
    ) -> None:
        """An appropriate error is given if the given image file does not exist."""
        runner = CliRunner(mix_stderr=False)
        does_not_exist_file = tmp_path / uuid.uuid4().hex
        commands: list[str] = [
            str(does_not_exist_file),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        assert result.stdout == ""
        expected_stderr = dedent(
            f"""\
            Usage: vuforia-cloud-reco [OPTIONS] IMAGE
            Try 'vuforia-cloud-reco --help' for help.

            Error: Invalid value for 'IMAGE': File '{does_not_exist_file}' does not exist.
            """,  # noqa: E501
        )
        assert result.stderr == expected_stderr


def test_version() -> None:
    """``vuforia-cloud-reco --version`` shows the version."""
    runner = CliRunner(mix_stderr=False)
    commands = ["--version"]
    result = runner.invoke(
        vuforia_cloud_reco,
        commands,
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.stdout.startswith("vuforia-cloud-reco, version ")


class TestMaxNumResults:
    """Tests for the ``--max-num-results`` option."""

    @staticmethod
    def test_default(
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """By default the maximum number of results is 1."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_2 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        vws_client.wait_for_target_processed(target_id=target_id_2)

        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        result_data = yaml.safe_load(result.stdout)
        assert len(result_data) == 1

    @staticmethod
    def test_custom(
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """It is possible to set a custom ``--max-num-results``."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_2 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_3 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        vws_client.wait_for_target_processed(target_id=target_id_2)
        vws_client.wait_for_target_processed(target_id=target_id_3)

        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        max_num_results = 2
        commands = [
            str(new_file),
            "--max-num-results",
            str(max_num_results),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        result_data = yaml.safe_load(result.stdout)
        assert len(result_data) == max_num_results

    @staticmethod
    def test_out_of_range(
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """``--max-num-results`` must be between 1 and 50."""
        runner = CliRunner(mix_stderr=False)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--max-num-results",
            str(0),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        expected_stderr_substring = (
            "Error: Invalid value for '--max-num-results': 0 is not in the "
            "range 1<=x<=50."
        )
        assert expected_stderr_substring in result.stderr


class TestIncludeTargetData:
    """Tests for the ``--include-target-data`` option."""

    @staticmethod
    def test_default(
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """By default, target data is only returned in the top match."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_2 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        vws_client.wait_for_target_processed(target_id=target_id_2)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--max-num-results",
            str(2),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        matches = yaml.safe_load(result.stdout)
        top_match, second_match = matches
        assert top_match["target_data"] is not None
        assert second_match["target_data"] is None

    @staticmethod
    def test_top(
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """When 'top' is given, target data is only returned in the top match."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_2 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        vws_client.wait_for_target_processed(target_id=target_id_2)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--max-num-results",
            str(2),
            "--include-target-data",
            "top",
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        matches = yaml.safe_load(result.stdout)
        top_match, second_match = matches
        assert top_match["target_data"] is not None
        assert second_match["target_data"] is None

    @staticmethod
    def test_none(
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """When 'none' is given, target data is not returned in any match."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_2 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        vws_client.wait_for_target_processed(target_id=target_id_2)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--max-num-results",
            str(2),
            "--include-target-data",
            "none",
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        matches = yaml.safe_load(result.stdout)
        top_match, second_match = matches
        assert top_match["target_data"] is None
        assert second_match["target_data"] is None

    @staticmethod
    def test_all(
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """When 'all' is given, target data is returned in all matches."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        target_id_2 = vws_client.add_target(
            name=uuid.uuid4().hex,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        vws_client.wait_for_target_processed(target_id=target_id_2)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)

        commands = [
            str(new_file),
            "--max-num-results",
            str(2),
            "--include-target-data",
            "all",
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        matches = yaml.safe_load(result.stdout)
        top_match, second_match = matches
        assert top_match["target_data"] is not None
        assert second_match["target_data"] is not None

    @staticmethod
    def test_other(
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        mock_database: VuforiaDatabase,
    ) -> None:
        """When a string other than 'top', 'all', or 'none' is given, an error is
        shown.
        """
        runner = CliRunner(mix_stderr=False)
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            str(new_file),
            "--max-num-results",
            str(2),
            "--include-target-data",
            "other",
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        expected_stderr = (
            "'--include-target-data': 'other' is not one of 'top', 'none', "
            "'all'."
        )
        assert expected_stderr in result.stderr


def test_base_vwq_url(high_quality_image: io.BytesIO, tmp_path: Path) -> None:
    """It is possible to use query a target to a database under a custom VWQ
    URL.
    """
    runner = CliRunner(mix_stderr=False)
    base_vwq_url = "http://example.com"
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    with MockVWS(base_vwq_url=base_vwq_url) as mock:
        mock_database = VuforiaDatabase()
        mock.add_database(database=mock_database)
        vws_client = VWS(
            server_access_key=mock_database.server_access_key,
            server_secret_key=mock_database.server_secret_key,
        )

        target_id = vws_client.add_target(
            name="x",
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )

        vws_client.wait_for_target_processed(target_id=target_id)

        commands = [
            str(new_file),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
            "--base-vwq-url",
            base_vwq_url,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )
        assert result.exit_code == 0

    [match] = yaml.safe_load(result.stdout)
    assert match["target_id"] == target_id


def test_env_var_credentials(
    high_quality_image: io.BytesIO,
    tmp_path: Path,
    mock_database: VuforiaDatabase,
) -> None:
    """It is possible to use environment variables to set the credentials."""
    runner = CliRunner(mix_stderr=False)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    commands = [str(new_file)]
    result = runner.invoke(
        vuforia_cloud_reco,
        commands,
        catch_exceptions=False,
        env={
            "VUFORIA_CLIENT_ACCESS_KEY": mock_database.client_access_key,
            "VUFORIA_CLIENT_SECRET_KEY": mock_database.client_secret_key,
        },
    )
    assert result.exit_code == 0



"""Tests for how errors from VWS are handled by the CLI."""

import io
import uuid
from pathlib import Path

from click.testing import CliRunner
from freezegun import freeze_time
from mock_vws import MockVWS
from mock_vws.database import VuforiaDatabase
from mock_vws.states import States
from vws import VWS
from vws_cli import vws_group


def test_target_id_does_not_exist(mock_database: VuforiaDatabase) -> None:
    """Commands which take a target ID show an error if that does not map to a
    target in the database.
    """
    runner = CliRunner(mix_stderr=False)
    for command_name, command in vws_group.commands.items():
        if "target_id" in [option.name for option in command.params]:
            args = [
                command_name,
                "--target-id",
                "abc12345",
                "--server-access-key",
                mock_database.server_access_key,
                "--server-secret-key",
                mock_database.server_secret_key,
            ]
            result = runner.invoke(vws_group, args, catch_exceptions=False)
            assert result.exit_code == 1
            expected_stderr = 'Error: Target "abc12345" does not exist.\n'
            assert result.stderr == expected_stderr
            assert result.stdout == ""


def test_bad_image(
    mock_database: VuforiaDatabase,
    tmp_path: Path,
) -> None:
    """An error is given when Vuforia returns a ``BadImage`` error. For example,
    when a corrupt image is uploaded.
    """
    new_file = tmp_path / uuid.uuid4().hex
    new_file.write_bytes(data=b"Not an image")
    runner = CliRunner(mix_stderr=False)
    args = [
        "add-target",
        "--name",
        "x",
        "--width",
        "0.1",
        "--image",
        str(new_file),
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, args, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = (
        "Error: The given image is corrupted or the format is not supported.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_fail_bad_request(
    mock_database: VuforiaDatabase,
    high_quality_image: io.BytesIO,
    tmp_path: Path,
) -> None:
    """An error is given when Vuforia returns a ``Fail`` error with a ``400``
    error code. For example, when the given server access key does not exist.

    With ``vws_python`` we cannot get a (guaranteed) 500 error or 422 response
    with a ``Fail`` error.
    """
    new_file = tmp_path / uuid.uuid4().hex
    new_file.write_bytes(data=high_quality_image.getvalue())
    runner = CliRunner(mix_stderr=False)
    args = [
        "add-target",
        "--name",
        "x",
        "--width",
        "0.1",
        "--image",
        str(new_file),
        "--server-access-key",
        "does_not_exist_key",
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, args, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = (
        "Error: The request made to Vuforia was invalid and could not be "
        "processed. "
        "Check the given parameters.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_metadata_too_large(
    mock_database: VuforiaDatabase,
    high_quality_image: io.BytesIO,
    tmp_path: Path,
) -> None:
    """An error is given when the given metadata is too large."""
    new_file = tmp_path / uuid.uuid4().hex
    new_file.write_bytes(data=high_quality_image.getvalue())
    runner = CliRunner(mix_stderr=False)
    args = [
        "add-target",
        "--name",
        "x",
        "--width",
        "0.1",
        "--image",
        str(new_file),
        "--application-metadata",
        "a" * 1024 * 1024 * 10,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, args, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = "Error: The given metadata is too large.\n"
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_image_too_large(
    mock_database: VuforiaDatabase,
    png_too_large: io.BytesIO,
    tmp_path: Path,
) -> None:
    """An error is given when the given image is too large."""
    runner = CliRunner(mix_stderr=False)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = png_too_large.getvalue()
    new_file.write_bytes(data=image_data)
    commands = [
        "add-target",
        "--name",
        "foo",
        "--width",
        "0.1",
        "--image",
        str(new_file),
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = "Error: The given image is too large.\n"
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_target_name_exist(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
    tmp_path: Path,
) -> None:
    """An error is given when there is already a target with the given name."""
    name = "foobar"
    vws_client.add_target(
        name=name,
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )

    runner = CliRunner(mix_stderr=False)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    commands = [
        "add-target",
        "--name",
        name,
        "--width",
        "0.1",
        "--image",
        str(new_file),
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = 'Error: There is already a target named "foobar".\n'
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_project_inactive(
    high_quality_image: io.BytesIO,
    tmp_path: Path,
) -> None:
    """An error is given if the project is inactive and the desired action cannot
    be taken because of this.
    """
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    database = VuforiaDatabase(state=States.PROJECT_INACTIVE)
    with MockVWS() as mock:
        mock.add_database(database=database)
        runner = CliRunner(mix_stderr=False)
        commands = [
            "add-target",
            "--name",
            "foo",
            "--width",
            "0.1",
            "--image",
            str(new_file),
            "--server-access-key",
            database.server_access_key,
            "--server-secret-key",
            database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)

    assert result.exit_code == 1
    expected_stderr = (
        "Error: The project associated with the given keys is inactive.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_unknown_vws_error(
    mock_database: VuforiaDatabase,
    high_quality_image: io.BytesIO,
    tmp_path: Path,
) -> None:
    """When an unknown VWS error is given, e.g. what is given when some bad names
    are given, an error is given.
    """
    runner = CliRunner(mix_stderr=False)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    max_char_value = 65535
    bad_name = chr(max_char_value + 1)

    commands = [
        "add-target",
        "--name",
        bad_name,
        "--width",
        "0.1",
        "--image",
        str(new_file),
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = (
        "Error: There was an unknown error from Vuforia. "
        "This may be because there is a problem with the given name.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_target_status_processing(
    vws_client: VWS,
    high_quality_image: io.BytesIO,
    mock_database: VuforiaDatabase,
) -> None:
    """An error is given when trying to delete a target which is processing."""
    runner = CliRunner(mix_stderr=False)

    target_id = vws_client.add_target(
        name="x",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )

    commands = [
        "delete-target",
        "--target-id",
        target_id,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]

    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = (
        f'Error: The target "{target_id}" cannot be deleted as it is in the '
        "processing state.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_target_status_not_success(
    vws_client: VWS,
    high_quality_image: io.BytesIO,
    mock_database: VuforiaDatabase,
) -> None:
    """An error is given when updating a target which has a status
    which is not "Success".
    """
    runner = CliRunner(mix_stderr=False)
    target_id = vws_client.add_target(
        name="x",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )

    commands = [
        "update-target",
        "--target-id",
        target_id,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]

    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = (
        f'Error: The target "{target_id}" cannot be updated as it is in the '
        "processing state.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_authentication_failure(mock_database: VuforiaDatabase) -> None:
    """An error is given when the secret key is incorrect."""
    runner = CliRunner(mix_stderr=False)
    commands = [
        "list-targets",
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        "wrong_key",
    ]

    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 1
    expected_stderr = "The given secret key was incorrect.\n"
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_request_time_too_skewed(mock_database: VuforiaDatabase) -> None:
    """An error is given when the request time is more than 5 minutes different
    from the server time.
    """
    runner = CliRunner(mix_stderr=False)
    vws_max_time_skew = 60 * 5
    leeway = 10
    time_difference_from_now = vws_max_time_skew + leeway

    # We use a custom tick because we expect the following:
    #
    # * At least one time check when creating the request
    # * At least one time check when processing the request
    #
    # >= 1 ticks are acceptable.
    with freeze_time(auto_tick_seconds=time_difference_from_now):
        commands = [
            "list-targets",
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)

    expected_stderr = (
        "Error: Vuforia reported that the time given with this request was "
        "outside the expected range. "
        "This may be because the system clock is out of sync.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""



"""A CLI for the Vuforia Cloud Recognition Service API."""

from __future__ import annotations

import dataclasses
import io
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any

import click
import wrapt
import yaml
from vws import CloudRecoService
from vws.exceptions.cloud_reco_exceptions import (
    AuthenticationFailure,
    BadImage,
    InactiveProject,
    RequestTimeTooSkewed,
)
from vws.exceptions.custom_exceptions import (
    ActiveMatchingTargetsDeleteProcessing,
    RequestEntityTooLarge,
)
from vws.include_target_data import CloudRecoIncludeTargetData

from vws_cli import __version__
from vws_cli.options.credentials import (
    client_access_key_option,
    client_secret_key_option,
)


@wrapt.decorator  # type: ignore[misc]
def handle_vwq_exceptions(
    wrapped: Callable[..., str],
    instance: None,
    args: tuple[Any],
    kwargs: dict[Any, Any],
) -> None:
    """Show error messages and catch exceptions from ``VWS-Python``."""
    assert not instance  # This is to satisfy the "vulture" linter.
    error_message = ""

    try:
        wrapped(*args, **kwargs)
    except BadImage:
        error_message = (
            "Error: The given image is corrupted or the format is not "
            "supported."
        )
    except InactiveProject:
        error_message = (
            "Error: The project associated with the given keys is inactive."
        )
    except AuthenticationFailure:
        error_message = "The given secret key was incorrect."
    except RequestTimeTooSkewed:
        error_message = (
            "Error: Vuforia reported that the time given with this request "
            "was outside the expected range. "
            "This may be because the system clock is out of sync."
        )
    except RequestEntityTooLarge:
        error_message = "Error: The given image is too large."
    except ActiveMatchingTargetsDeleteProcessing:
        error_message = (
            "Error: The given image matches a target which was recently "
            "deleted."
        )
    else:
        return

    click.echo(error_message, err=True)
    sys.exit(1)


def image_argument(command: Callable[..., None]) -> Callable[..., None]:
    """An argument decorator for choosing a query image."""
    click_argument_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.argument(
        "image",
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=False,
            path_type=Path,
        ),
    )

    function: Callable[..., None] = click_argument_function(command)
    return function


def max_num_results_option(
    command: Callable[..., None],
) -> Callable[..., None]:
    """An option decorator for choosing the maximum number of query results."""
    maximum = 50
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--max-num-results",
        type=click.IntRange(min=1, max=maximum),
        default=1,
        help=(
            "The maximum number of matching targets to be returned. "
            f"Must be <= {maximum}."
        ),
        show_default=True,
    )

    function: Callable[..., None] = click_option_function(command)
    return function


def include_target_data_callback(
    ctx: click.core.Context,
    param: click.core.Option | click.core.Parameter,
    value: str,
) -> CloudRecoIncludeTargetData:
    """Use as a callback for active flag options."""
    # This is to satisfy the "vulture" linter.
    assert ctx
    assert param

    return {
        "top": CloudRecoIncludeTargetData.TOP,
        "none": CloudRecoIncludeTargetData.NONE,
        "all": CloudRecoIncludeTargetData.ALL,
    }[value]


def include_target_data_option(
    command: Callable[..., None],
) -> Callable[..., None]:
    """An option decorator for choosing whether to include target data."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--include-target-data",
        type=click.Choice(["top", "none", "all"], case_sensitive=True),
        default="top",
        callback=include_target_data_callback,
        help=(
            "Whether target_data records shall be returned for the matched "
            "targets. Accepted values are top (default value, only return "
            "target_data for top ranked match), none (return no target_data), "
            "all (for all matched targets)."
        ),
        show_default=True,
    )

    function: Callable[..., None] = click_option_function(command)
    return function


def base_vwq_url_option(command: Callable[..., None]) -> Callable[..., None]:
    """An option decorator for choosing the base VWQ URL."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--base-vwq-url",
        type=click.STRING,
        default="https://cloudreco.vuforia.com",
        help="The base URL for the VWQ API.",
        show_default=True,
    )

    function: Callable[..., None] = click_option_function(command)
    return function


@click.command(name="vuforia-cloud-reco")
@image_argument
@client_access_key_option
@client_secret_key_option
@include_target_data_option
@max_num_results_option
@base_vwq_url_option
@handle_vwq_exceptions  # type: ignore[misc]
# We set the ``version`` parameter because in PyInstaller binaries,
# ``pkg_resources`` is not available.
#
# Click uses ``pkg_resources`` to determine the version if it is not given.
@click.version_option(version=__version__)
def vuforia_cloud_reco(
    image: Path,
    client_access_key: str,
    client_secret_key: str,
    max_num_results: int,
    include_target_data: CloudRecoIncludeTargetData,
    base_vwq_url: str,
) -> None:
    """Make a request to the Vuforia Cloud Recognition Service API."""
    client = CloudRecoService(
        client_access_key=client_access_key,
        client_secret_key=client_secret_key,
        base_vwq_url=base_vwq_url,
    )
    query_result = client.query(
        image=io.BytesIO(image.read_bytes()),
        max_num_results=max_num_results,
        include_target_data=include_target_data,
    )
    query_result_dict_list = [dataclasses.asdict(res) for res in query_result]
    yaml_list = yaml.dump(query_result_dict_list)
    click.echo(yaml_list)



"""``click`` options regarding credentials."""

from __future__ import annotations

from collections.abc import Callable

import click


def server_access_key_option(
    command: Callable[..., None] | None = None,
) -> Callable[..., None]:
    """An option decorator for the Vuforia server access key."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--server-access-key",
        type=str,
        help=(
            "A Vuforia server access key to use to access the Vuforia Web "
            "Services API."
        ),
        required=True,
        envvar="VUFORIA_SERVER_ACCESS_KEY",
        show_envvar=True,
    )
    assert command is not None
    function: Callable[..., None] = click_option_function(command)
    return function


def server_secret_key_option(
    command: Callable[..., None] | None = None,
) -> Callable[..., None]:
    """An option decorator for the Vuforia server secret key."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--server-secret-key",
        type=str,
        help=(
            "A Vuforia server secret key to use to access the Vuforia Web "
            "Services API."
        ),
        required=True,
        envvar="VUFORIA_SERVER_SECRET_KEY",
        show_envvar=True,
    )
    assert command is not None
    function: Callable[..., None] = click_option_function(command)
    return function


def client_access_key_option(
    command: Callable[..., None] | None = None,
) -> Callable[..., None]:
    """An option decorator for the Vuforia client access key."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--client-access-key",
        type=str,
        help=(
            "A Vuforia client access key to use to access the Vuforia "
            "Cloud Recognition API."
        ),
        required=True,
        envvar="VUFORIA_CLIENT_ACCESS_KEY",
        show_envvar=True,
    )
    assert command is not None
    function: Callable[..., None] = click_option_function(command)
    return function


def client_secret_key_option(
    command: Callable[..., None] | None = None,
) -> Callable[..., None]:
    """An option decorator for the Vuforia client secret key."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--client-secret-key",
        type=str,
        help=(
            "A Vuforia client secret key to use to access the Vuforia "
            "Cloud Recognition API."
        ),
        required=True,
        envvar="VUFORIA_CLIENT_SECRET_KEY",
        show_envvar=True,
    )
    assert command is not None
    function: Callable[..., None] = click_option_function(command)
    return function



"""Tests for the VWS CLI."""

from click.testing import CliRunner
from vws_cli import vws_group


def test_version() -> None:
    """The CLI version is shown with ``vws --version``."""
    runner = CliRunner()
    result = runner.invoke(vws_group, ["--version"], catch_exceptions=False)
    assert result.exit_code == 0
    expected = "vws, version"
    assert expected in result.stdout



"""Tests for how errors from the Cloud Reco Service are handled by the CLI."""

from __future__ import annotations

import io
import time
import uuid
from pathlib import Path

from click.testing import CliRunner
from freezegun import freeze_time
from mock_vws import MockVWS
from mock_vws.database import VuforiaDatabase
from mock_vws.states import States
from vws import VWS
from vws_cli.query import vuforia_cloud_reco


def test_authentication_failure(
    mock_database: VuforiaDatabase,
    tmp_path: Path,
    high_quality_image: io.BytesIO,
) -> None:
    """An error is given when the secret key is incorrect."""
    runner = CliRunner(mix_stderr=False)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    commands: list[str] = [
        str(new_file),
        "--client-access-key",
        mock_database.client_access_key,
        "--client-secret-key",
        "wrong_secret_key",
    ]
    result = runner.invoke(
        vuforia_cloud_reco,
        commands,
        catch_exceptions=False,
    )
    expected_stderr = "The given secret key was incorrect.\n"
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_image_too_large(
    mock_database: VuforiaDatabase,
    tmp_path: Path,
    png_too_large: io.BytesIO,
) -> None:
    """An error is given when the image is too large."""
    runner = CliRunner(mix_stderr=False)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = png_too_large.getvalue()
    new_file.write_bytes(data=image_data)
    commands: list[str] = [
        str(new_file),
        "--client-access-key",
        mock_database.client_access_key,
        "--client-secret-key",
        mock_database.client_secret_key,
    ]
    result = runner.invoke(
        vuforia_cloud_reco,
        commands,
        catch_exceptions=False,
    )
    expected_stderr = "Error: The given image is too large.\n"
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_active_matching_targets_delete_processing(
    mock_database: VuforiaDatabase,
    tmp_path: Path,
    high_quality_image: io.BytesIO,
    vws_client: VWS,
) -> None:
    """An error is given when the image matches a target which has recently been
    deleted.
    """
    runner = CliRunner(mix_stderr=False)

    target_id = vws_client.add_target(
        name="x",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )
    vws_client.wait_for_target_processed(target_id=target_id)
    vws_client.delete_target(target_id=target_id)
    time.sleep(0.2)
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    commands: list[str] = [
        str(new_file),
        "--client-access-key",
        mock_database.client_access_key,
        "--client-secret-key",
        mock_database.client_secret_key,
    ]
    result = runner.invoke(
        vuforia_cloud_reco,
        commands,
        catch_exceptions=False,
    )
    expected_stderr = (
        "Error: The given image matches a target which was recently deleted.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_bad_image(
    mock_database: VuforiaDatabase,
    tmp_path: Path,
) -> None:
    """An error is given when Vuforia returns a ``BadImage`` error. For example,
    when a corrupt image is uploaded.
    """
    new_file = tmp_path / uuid.uuid4().hex
    new_file.write_bytes(data=b"Not an image")
    runner = CliRunner(mix_stderr=False)
    commands: list[str] = [
        str(new_file),
        "--client-access-key",
        mock_database.client_access_key,
        "--client-secret-key",
        mock_database.client_secret_key,
    ]
    result = runner.invoke(
        vuforia_cloud_reco,
        commands,
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    expected_stderr = (
        "Error: The given image is corrupted or the format is not supported.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_inactive_project(
    high_quality_image: io.BytesIO,
    tmp_path: Path,
) -> None:
    """An error is given if the project is inactive and the desired action cannot
    be taken because of this.
    """
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)
    database = VuforiaDatabase(state=States.PROJECT_INACTIVE)
    with MockVWS() as mock:
        mock.add_database(database=database)
        runner = CliRunner(mix_stderr=False)
        commands = [
            str(new_file),
            "--client-access-key",
            database.client_access_key,
            "--client-secret-key",
            database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )

    assert result.exit_code == 1
    expected_stderr = (
        "Error: The project associated with the given keys is inactive.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""


def test_request_time_too_skewed(
    high_quality_image: io.BytesIO,
    mock_database: VuforiaDatabase,
    tmp_path: Path,
) -> None:
    """An error is given when the request time is more than 65 minutes different
    from the server time.
    """
    runner = CliRunner(mix_stderr=False)
    vwq_max_time_skew = 60 * 65
    leeway = 10
    time_difference_from_now = vwq_max_time_skew + leeway
    new_file = tmp_path / uuid.uuid4().hex
    image_data = high_quality_image.getvalue()
    new_file.write_bytes(data=image_data)

    # We use a custom tick because we expect the following:
    #
    # * At least one time check when creating the request
    # * At least one time check when processing the request
    #
    # >= 1 ticks are acceptable.
    with freeze_time(auto_tick_seconds=time_difference_from_now):
        commands = [
            str(new_file),
            "--client-access-key",
            mock_database.client_access_key,
            "--client-secret-key",
            mock_database.client_secret_key,
        ]
        result = runner.invoke(
            vuforia_cloud_reco,
            commands,
            catch_exceptions=False,
        )

    expected_stderr = (
        "Error: Vuforia reported that the time given with this request was "
        "outside the expected range. "
        "This may be because the system clock is out of sync.\n"
    )
    assert result.stderr == expected_stderr
    assert result.stdout == ""



# pylint:disable=too-many-lines
"""Tests for VWS CLI commands."""

import base64
import io
import random
import uuid
from pathlib import Path
from textwrap import dedent

import pytest
import yaml
from click.testing import CliRunner
from freezegun import freeze_time
from mock_vws import MockVWS
from mock_vws.database import VuforiaDatabase
from vws import VWS, CloudRecoService
from vws.reports import TargetStatuses
from vws_cli import vws_group


def test_get_database_summary_report(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
) -> None:
    """It is possible to get a database summary report."""
    runner = CliRunner()
    for name in ("a", "b"):
        vws_client.add_target(
            name=name,
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )

    commands = [
        "get-database-summary-report",
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 0
    result_data = yaml.safe_load(result.stdout)
    expected_result_data = {
        "active_images": 0,
        "current_month_recos": 0,
        "failed_images": 0,
        "inactive_images": 0,
        "name": mock_database.database_name,
        "previous_month_recos": 0,
        "processing_images": 2,
        "reco_threshold": 1000,
        "request_quota": 100000,
        "request_usage": 0,
        "target_quota": 1000,
        "total_recos": 0,
    }
    assert result_data == expected_result_data


def test_list_targets(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
) -> None:
    """It is possible to get a list of targets in the database."""
    runner = CliRunner()
    target_id_1 = vws_client.add_target(
        name="x1",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )
    target_id_2 = vws_client.add_target(
        name="x2",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )
    commands = [
        "list-targets",
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 0
    result_data = yaml.safe_load(result.stdout)
    expected_result_data = [target_id_1, target_id_2]
    # We do not expect a particular order.
    assert sorted(result_data) == sorted(expected_result_data)


def test_get_target_record(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
) -> None:
    """It is possible to get a target record."""
    runner = CliRunner()
    target_id = vws_client.add_target(
        name="x",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )
    commands = [
        "get-target-record",
        "--target-id",
        target_id,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 0
    result_data = yaml.safe_load(result.stdout)
    expected_result_data = {
        "active_flag": True,
        "name": "x",
        "reco_rating": "",
        "target_id": target_id,
        "tracking_rating": -1,
        "width": 1,
    }
    assert result_data == expected_result_data


def test_get_target_summary_report(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
) -> None:
    """It is possible to get a target summary report."""
    runner = CliRunner()
    upload_date = "2015-04-29"
    with freeze_time(upload_date):
        target_id = vws_client.add_target(
            name="x",
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )

    commands = [
        "get-target-summary-report",
        "--target-id",
        target_id,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 0
    result_data = yaml.safe_load(result.stdout)
    expected_result_data = {
        "active_flag": True,
        "current_month_recos": 0,
        "database_name": mock_database.database_name,
        "previous_month_recos": 0,
        "status": "success",
        "target_name": "x",
        "total_recos": 0,
        "tracking_rating": result_data["tracking_rating"],
        "upload_date": upload_date,
    }
    assert result_data == expected_result_data


def test_delete_target(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
) -> None:
    """It is possible to delete a target."""
    runner = CliRunner()
    target_id = vws_client.add_target(
        name="x",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )
    assert vws_client.list_targets() == [target_id]
    commands = [
        "delete-target",
        "--target-id",
        target_id,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    vws_client.wait_for_target_processed(target_id=target_id)
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 0
    assert result.stdout == ""
    assert vws_client.list_targets() == []


def test_get_duplicate_targets(
    mock_database: VuforiaDatabase,
    vws_client: VWS,
    high_quality_image: io.BytesIO,
) -> None:
    """It is possible to get a list of duplicate targets."""
    runner = CliRunner()
    target_id = vws_client.add_target(
        name="x",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )
    target_id_2 = vws_client.add_target(
        name="x2",
        width=1,
        image=high_quality_image,
        active_flag=True,
        application_metadata=None,
    )

    vws_client.wait_for_target_processed(target_id)
    vws_client.wait_for_target_processed(target_id_2)

    commands = [
        "get-duplicate-targets",
        "--target-id",
        target_id,
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
    ]
    result = runner.invoke(vws_group, commands, catch_exceptions=False)
    assert result.exit_code == 0
    result_data = yaml.safe_load(result.stdout)
    expected_result_data = [target_id_2]
    assert result_data == expected_result_data


class TestAddTarget:
    """Tests for ``vws add-target``."""

    @staticmethod
    def test_add_target(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        cloud_reco_client: CloudRecoService,
    ) -> None:
        """It is possible to add a target."""
        runner = CliRunner()
        new_file = tmp_path / uuid.uuid4().hex
        name = uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        width = random.uniform(a=0.01, b=50)
        commands = [
            "add-target",
            "--name",
            name,
            "--width",
            str(width),
            "--image",
            str(new_file),
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0

        target_id = result.stdout.strip()
        target_details = vws_client.get_target_record(target_id=target_id)
        target_record = target_details.target_record
        assert target_record.name == name
        assert target_record.width == width
        assert target_record.active_flag is True
        vws_client.wait_for_target_processed(target_id=target_id)

        [query_result] = cloud_reco_client.query(image=high_quality_image)
        assert query_result.target_id == target_id
        target_data = query_result.target_data
        assert target_data is not None
        assert target_data.application_metadata is None

    @staticmethod
    def test_image_file_does_not_exist(
        mock_database: VuforiaDatabase,
        tmp_path: Path,
    ) -> None:
        """An appropriate error is given if the given image file does not exist."""
        runner = CliRunner(mix_stderr=False)
        does_not_exist_file = tmp_path / uuid.uuid4().hex
        commands = [
            "add-target",
            "--name",
            "foo",
            "--width",
            "1",
            "--image",
            str(does_not_exist_file),
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        assert result.stdout == ""
        expected_stderr = dedent(
            f"""\
            Usage: vws add-target [OPTIONS]
            Try 'vws add-target -h' for help.

            Error: Invalid value for '--image': File '{does_not_exist_file}' does not exist.
            """,  # noqa: E501
        )
        assert result.stderr == expected_stderr

    @staticmethod
    def test_image_file_is_dir(
        mock_database: VuforiaDatabase,
        tmp_path: Path,
    ) -> None:
        """An appropriate error is given if the given image file path points to a
        directory.
        """
        runner = CliRunner(mix_stderr=False)
        commands = [
            "add-target",
            "--name",
            "foo",
            "--width",
            "1",
            "--image",
            str(tmp_path),
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        assert result.stdout == ""
        expected_stderr = dedent(
            f"""\
            Usage: vws add-target [OPTIONS]
            Try 'vws add-target -h' for help.

            Error: Invalid value for '--image': File '{tmp_path}' is a directory.
            """,
        )
        assert result.stderr == expected_stderr

    @staticmethod
    def test_relative_path(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
    ) -> None:
        """Image file paths are resolved."""
        runner = CliRunner(mix_stderr=False)
        new_filename = uuid.uuid4().hex
        original_image_file = tmp_path / "foo"
        image_data = high_quality_image.getvalue()
        original_image_file.write_bytes(image_data)
        name = uuid.uuid4().hex
        commands = [
            "add-target",
            "--name",
            name,
            "--width",
            "1",
            "--image",
            new_filename,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        with runner.isolated_filesystem():
            new_file = Path(new_filename)
            new_file.symlink_to(original_image_file)
            result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0
        target_id = result.stdout.strip()
        target_record = vws_client.get_target_record(target_id=target_id)
        assert target_record.target_record.name == name

    @staticmethod
    def test_custom_metadata(
        mock_database: VuforiaDatabase,
        cloud_reco_client: CloudRecoService,
        vws_client: VWS,
        tmp_path: Path,
        high_quality_image: io.BytesIO,
    ) -> None:
        """Custom metadata can be given."""
        runner = CliRunner()
        new_file = tmp_path / uuid.uuid4().hex
        name = uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        application_metadata = uuid.uuid4().hex
        metadata_bytes = application_metadata.encode("ascii")
        base64_encoded_metadata_bytes = base64.b64encode(metadata_bytes)
        base64_encoded_metadata = base64_encoded_metadata_bytes.decode("ascii")
        commands = [
            "add-target",
            "--name",
            name,
            "--width",
            "0.1",
            "--image",
            str(new_file),
            "--application-metadata",
            base64_encoded_metadata,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0
        target_id = result.stdout.strip()
        vws_client.wait_for_target_processed(target_id=target_id)
        [query_result] = cloud_reco_client.query(image=high_quality_image)
        assert query_result.target_id == target_id
        target_data = query_result.target_data
        assert target_data is not None
        assert target_data.application_metadata == base64_encoded_metadata

    @staticmethod
    @pytest.mark.parametrize(
        ("active_flag_given", "active_flag_expected"),
        [
            ("true", True),
            ("false", False),
        ],
    )
    def test_custom_active_flag(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        active_flag_given: str,
        *,
        active_flag_expected: bool,
    ) -> None:
        """The Active Flag of the new target can be chosen."""
        runner = CliRunner()
        new_file = tmp_path / uuid.uuid4().hex
        image_data = high_quality_image.getvalue()
        new_file.write_bytes(data=image_data)
        commands = [
            "add-target",
            "--name",
            "foo",
            "--width",
            "0.1",
            "--image",
            str(new_file),
            "--active-flag",
            active_flag_given,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0

        target_id = result.stdout.strip()
        target_details = vws_client.get_target_record(target_id=target_id)
        active_flag = target_details.target_record.active_flag
        assert active_flag is active_flag_expected


class TestWaitForTargetProcessed:
    """Tests for ``vws wait-for-target-processed``."""

    @staticmethod
    def test_wait_for_target_processed(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
    ) -> None:
        """It is possible to use a command to wait for a target to be processed."""
        runner = CliRunner()
        target_id = vws_client.add_target(
            name="x",
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        report = vws_client.get_target_summary_report(target_id=target_id)
        assert report.status == TargetStatuses.PROCESSING
        commands = [
            "wait-for-target-processed",
            "--target-id",
            target_id,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0
        assert result.stdout == ""
        report = vws_client.get_target_summary_report(target_id=target_id)
        assert report.status != TargetStatuses.PROCESSING

    @staticmethod
    def test_default_seconds_between_requests(
        high_quality_image: io.BytesIO,
    ) -> None:
        """By default, 0.2 seconds are waited between polling requests."""
        runner = CliRunner()
        with MockVWS(processing_time_seconds=0.5) as mock:
            mock_database = VuforiaDatabase()
            mock.add_database(database=mock_database)
            vws_client = VWS(
                server_access_key=mock_database.server_access_key,
                server_secret_key=mock_database.server_secret_key,
            )

            target_id = vws_client.add_target(
                name="x",
                width=1,
                image=high_quality_image,
                active_flag=True,
                application_metadata=None,
            )

            commands = [
                "wait-for-target-processed",
                "--target-id",
                target_id,
                "--server-access-key",
                mock_database.server_access_key,
                "--server-secret-key",
                mock_database.server_secret_key,
            ]
            result = runner.invoke(vws_group, commands, catch_exceptions=False)
            assert result.exit_code == 0
            assert result.stdout == ""
            report = vws_client.get_database_summary_report()
            expected_requests = (
                # Add target request
                1
                +
                # Database summary request
                1
                +
                # Initial request
                1
                +
                # Request after 0.2 seconds - not processed
                1
                +
                # Request after 0.4 seconds - not processed
                # This assumes that there is less than 0.1 seconds taken
                # between the start of the target processing and the start of
                # waiting for the target to be processed.
                1
                +
                # Request after 0.6 seconds - processed
                1
            )
            # At the time of writing there is a bug which prevents request
            # usage from being tracked so we cannot track this.
            expected_requests = 0
            assert report.request_usage == expected_requests

    @staticmethod
    def test_custom_seconds_between_requests(
        high_quality_image: io.BytesIO,
    ) -> None:
        """It is possible to customize the time waited between polling requests."""
        runner = CliRunner()
        with MockVWS(processing_time_seconds=0.5) as mock:
            mock_database = VuforiaDatabase()
            mock.add_database(database=mock_database)
            vws_client = VWS(
                server_access_key=mock_database.server_access_key,
                server_secret_key=mock_database.server_secret_key,
            )

            target_id = vws_client.add_target(
                name="x",
                width=1,
                image=high_quality_image,
                active_flag=True,
                application_metadata=None,
            )

            commands = [
                "wait-for-target-processed",
                "--target-id",
                target_id,
                "--seconds-between-requests",
                "0.3",
                "--server-access-key",
                mock_database.server_access_key,
                "--server-secret-key",
                mock_database.server_secret_key,
            ]
            result = runner.invoke(vws_group, commands, catch_exceptions=False)
            assert result.exit_code == 0
            assert result.stdout == ""
            report = vws_client.get_database_summary_report()
            expected_requests = (
                # Add target request
                1
                +
                # Database summary request
                1
                +
                # Initial request
                1
                +
                # Request after 0.3 seconds - not processed
                # This assumes that there is less than 0.2 seconds taken
                # between the start of the target processing and the start of
                # waiting for the target to be processed.
                1
                +
                # Request after 0.6 seconds - processed
                1
            )
            # At the time of writing there is a bug which prevents request
            # usage from being tracked so we cannot track this.
            expected_requests = 0
            assert report.request_usage == expected_requests

    @staticmethod
    def test_custom_seconds_too_small(mock_database: VuforiaDatabase) -> None:
        """The minimum valid value for ``--seconds-between-requests`` is 0.05
        seconds.
        """
        runner = CliRunner(mix_stderr=False)
        commands = [
            "wait-for-target-processed",
            "--target-id",
            "x",
            "--seconds-between-requests",
            "0.01",
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code != 0
        assert result.stdout == ""
        expected_substring = "0.01 is not in the range x>=0.05."
        assert expected_substring in result.stderr

    @staticmethod
    def test_custom_timeout(high_quality_image: io.BytesIO) -> None:
        """It is possible to set a maximum timeout."""
        runner = CliRunner(mix_stderr=False)
        with MockVWS(processing_time_seconds=0.5) as mock:
            mock_database = VuforiaDatabase()
            mock.add_database(database=mock_database)
            vws_client = VWS(
                server_access_key=mock_database.server_access_key,
                server_secret_key=mock_database.server_secret_key,
            )

            target_id = vws_client.add_target(
                name="x",
                width=1,
                image=high_quality_image,
                active_flag=True,
                application_metadata=None,
            )

            report = vws_client.get_target_summary_report(target_id=target_id)
            assert report.status == TargetStatuses.PROCESSING

            commands = [
                "wait-for-target-processed",
                "--target-id",
                target_id,
                "--timeout-seconds",
                "0.1",
                "--server-access-key",
                mock_database.server_access_key,
                "--server-secret-key",
                mock_database.server_secret_key,
            ]
            result = runner.invoke(vws_group, commands, catch_exceptions=False)
            assert result.exit_code != 0
            assert result.stderr == "Timeout of 0.1 seconds reached.\n"

            commands = [
                "wait-for-target-processed",
                "--target-id",
                target_id,
                "--timeout-seconds",
                "0.5",
                "--server-access-key",
                mock_database.server_access_key,
                "--server-secret-key",
                mock_database.server_secret_key,
            ]
            result = runner.invoke(vws_group, commands, catch_exceptions=False)
            assert result.exit_code == 0
            report = vws_client.get_target_summary_report(target_id=target_id)
            assert report.status != TargetStatuses.PROCESSING

    @staticmethod
    def test_custom_timeout_too_small(mock_database: VuforiaDatabase) -> None:
        """The minimum valid value for ``--timeout-seconds`` is 0.05 seconds."""
        runner = CliRunner(mix_stderr=False)
        commands = [
            "wait-for-target-processed",
            "--target-id",
            "x",
            "--timeout-seconds",
            "0.01",
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code != 0
        expected_substring = "0.01 is not in the range x>=0.05."
        assert expected_substring in result.stderr


class TestUpdateTarget:
    """Tests for ``vws update-target``."""

    @staticmethod
    def test_update_target(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
        cloud_reco_client: CloudRecoService,
        different_high_quality_image: io.BytesIO,
    ) -> None:
        """It is possible to update a target."""
        runner = CliRunner(mix_stderr=False)
        old_name = uuid.uuid4().hex
        old_width = random.uniform(a=0.01, b=50)
        target_id = vws_client.add_target(
            name=old_name,
            width=old_width,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        new_application_metadata = base64.b64encode(b"a").decode("ascii")
        new_name = uuid.uuid4().hex
        new_width = random.uniform(a=0.01, b=50)
        new_image_file = tmp_path / uuid.uuid4().hex
        new_image_data = different_high_quality_image.getvalue()
        new_image_file.write_bytes(data=new_image_data)

        commands = [
            "update-target",
            "--target-id",
            target_id,
            "--name",
            new_name,
            "--width",
            str(new_width),
            "--image",
            str(new_image_file),
            "--active-flag",
            "true",
            "--application-metadata",
            new_application_metadata,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0
        assert result.stdout == ""

        vws_client.wait_for_target_processed(target_id=target_id)
        [
            matching_target,
        ] = cloud_reco_client.query(image=different_high_quality_image)
        assert matching_target.target_id == target_id
        query_target_data = matching_target.target_data
        assert query_target_data is not None
        query_metadata = query_target_data.application_metadata
        assert query_metadata == new_application_metadata

        commands = [
            "update-target",
            "--target-id",
            target_id,
            "--active-flag",
            "false",
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0
        assert result.stdout == ""
        target_details = vws_client.get_target_record(target_id=target_id)
        target_record = target_details.target_record
        assert not target_record.active_flag
        assert target_record.name == new_name
        assert target_record.width == new_width
        assert not target_record.active_flag

    @staticmethod
    def test_no_fields_given(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
    ) -> None:
        """It is possible to give no update fields."""
        runner = CliRunner(mix_stderr=False)
        old_name = uuid.uuid4().hex
        old_width = random.uniform(a=0.01, b=50)
        target_id = vws_client.add_target(
            name=old_name,
            width=old_width,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)

        commands = [
            "update-target",
            "--target-id",
            target_id,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        assert result.exit_code == 0
        assert result.stdout == ""

    @staticmethod
    def test_image_file_does_not_exist(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
    ) -> None:
        """An appropriate error is given if the given image file does not exist."""
        target_id = vws_client.add_target(
            name="x",
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        runner = CliRunner(mix_stderr=False)
        does_not_exist_file = tmp_path / uuid.uuid4().hex
        commands = [
            "update-target",
            "--target-id",
            target_id,
            "--image",
            str(does_not_exist_file),
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        assert result.stdout == ""
        expected_stderr = dedent(
            f"""\
            Usage: vws update-target [OPTIONS]
            Try 'vws update-target -h' for help.

            Error: Invalid value for '--image': File '{does_not_exist_file}' does not exist.
            """,  # noqa: E501
        )
        assert result.stderr == expected_stderr

    @staticmethod
    def test_image_file_is_dir(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
    ) -> None:
        """An appropriate error is given if the given image file path points to a
        directory.
        """
        target_id = vws_client.add_target(
            name="x",
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        runner = CliRunner(mix_stderr=False)
        commands = [
            "update-target",
            "--target-id",
            target_id,
            "--image",
            str(tmp_path),
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]
        result = runner.invoke(vws_group, commands, catch_exceptions=False)
        expected_result_code = 2
        assert result.exit_code == expected_result_code
        assert result.stdout == ""
        expected_stderr = dedent(
            f"""\
            Usage: vws update-target [OPTIONS]
            Try 'vws update-target -h' for help.

            Error: Invalid value for '--image': File '{tmp_path}' is a directory.
            """,
        )
        assert result.stderr == expected_stderr

    @staticmethod
    def test_relative_path(
        mock_database: VuforiaDatabase,
        vws_client: VWS,
        high_quality_image: io.BytesIO,
        tmp_path: Path,
    ) -> None:
        """Image file paths are resolved."""
        runner = CliRunner(mix_stderr=False)
        target_id = vws_client.add_target(
            name="x",
            width=1,
            image=high_quality_image,
            active_flag=True,
            application_metadata=None,
        )
        vws_client.wait_for_target_processed(target_id=target_id)
        new_filename = uuid.uuid4().hex
        original_image_file = tmp_path / "foo"
        image_data = high_quality_image.getvalue()
        original_image_file.write_bytes(image_data)
        commands = [
            "update-target",
            "--target-id",
            target_id,
            "--image",
            new_filename,
            "--server-access-key",
            mock_database.server_access_key,
            "--server-secret-key",
            mock_database.server_secret_key,
        ]

        with runner.isolated_filesystem():
            new_file = Path(new_filename)
            new_file.symlink_to(original_image_file)
            result = runner.invoke(vws_group, commands, catch_exceptions=False)

        assert result.exit_code == 0


def test_custom_base_url() -> None:
    """It is possible to use the API to connect to a database under a custom VWS
    URL.
    """
    runner = CliRunner(mix_stderr=False)
    base_vws_url = "http://example.com"
    mock_database = VuforiaDatabase()
    commands = [
        "list-targets",
        "--server-access-key",
        mock_database.server_access_key,
        "--server-secret-key",
        mock_database.server_secret_key,
        "--base-vws-url",
        base_vws_url,
    ]
    with MockVWS(base_vws_url=base_vws_url) as mock:
        mock.add_database(database=mock_database)
        result = runner.invoke(vws_group, commands, catch_exceptions=False)

    assert result.exit_code == 0



"""Make PyInstaller binaries for the platform that this is being run on."""

import contextlib
import shutil
import subprocess
import sys
from pathlib import Path


def remove_existing_files(scripts: set[Path]) -> None:
    """Remove files created when building binaries.

    This is to stop interference with future builds.
    """
    dist_dir = Path(".") / "dist"
    build_dir = Path(".") / "build"

    with contextlib.suppress(FileNotFoundError):
        shutil.rmtree(path=str(dist_dir))

    with contextlib.suppress(FileNotFoundError):
        shutil.rmtree(path=str(build_dir))

    for script in scripts:
        path = Path(script.name + ".spec")
        with contextlib.suppress(FileNotFoundError):
            path.unlink()


def create_binary(script: Path, repo_root: Path) -> None:
    """Use PyInstaller to create a binary from a script.

    Args:
        script: The script to create a binary for.
        repo_root: The path to the root of the repository.
    """
    # MANIFEST.in describes files that must be available which are not
    # necessarily Python files.
    # These include e.g. Dockerfiles.
    # We still need to include these in the binary.
    datas = []
    manifest = repo_root / "MANIFEST.in"
    with manifest.open() as manifest_file:
        for line in manifest_file.readlines():
            # We do not yet have support for other MANIFEST types.
            is_include_line = line.startswith("include")
            is_recursive_include_line = line.startswith("recursive-include")
            assert is_include_line or is_recursive_include_line
            if line.startswith("recursive-include"):
                _, manifest_path, _ = line.split()
            else:
                _, manifest_path = line.split()
            if manifest_path.startswith("src/"):
                if Path(manifest_path).is_file():
                    parent = Path(manifest_path).parent
                    manifest_path = str(parent)

                src_path_length = len("src/")
                path_without_src = manifest_path[src_path_length:]
                data_item = (str(repo_root / manifest_path), path_without_src)
                datas.append(data_item)

    pyinstaller_command = [
        "pyinstaller",
        str(script.resolve()),
        "--onefile",
        "--name",
        script.name + "-" + sys.platform,
    ]
    for data in datas:
        source, destination = data
        add_data_command = ["--add-data", f"{source}:{destination}"]
        pyinstaller_command += add_data_command

    subprocess.check_output(args=pyinstaller_command)


def create_binaries() -> None:
    """Make PyInstaller binaries for the platform that this is being run on.

    All binaries will be created in ``./dist``.
    """
    repo_root = Path(__file__).resolve().parent.parent
    script_dir = repo_root / "bin"
    scripts = set(script_dir.iterdir())
    remove_existing_files(scripts=scripts)
    for script in scripts:
        create_binary(script=script, repo_root=repo_root)


if __name__ == "__main__":
    create_binaries()



"""``click`` commands the VWS CLI."""

from __future__ import annotations

import dataclasses
import io
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any

import click
import wrapt
import yaml
from vws import VWS
from vws.exceptions.custom_exceptions import (
    TargetProcessingTimeout,
    UnknownVWSErrorPossiblyBadName,
)
from vws.exceptions.vws_exceptions import (
    AuthenticationFailure,
    BadImage,
    DateRangeError,
    Fail,
    ImageTooLarge,
    MetadataTooLarge,
    ProjectHasNoAPIAccess,
    ProjectInactive,
    ProjectSuspended,
    RequestQuotaReached,
    RequestTimeTooSkewed,
    TargetNameExist,
    TargetQuotaReached,
    TargetStatusNotSuccess,
    TargetStatusProcessing,
    UnknownTarget,
)

from vws_cli.options.credentials import (
    server_access_key_option,
    server_secret_key_option,
)
from vws_cli.options.targets import (
    ActiveFlagChoice,
    active_flag_option,
    application_metadata_option,
    target_id_option,
    target_image_option,
    target_name_option,
    target_width_option,
)


@wrapt.decorator  # type: ignore[misc]
def handle_vws_exceptions(  # noqa: E501 pylint:disable=too-many-branches,too-many-statements
    wrapped: Callable[..., str],
    instance: None,
    args: tuple[Any],
    kwargs: dict[Any, Any],
) -> None:
    """Show error messages and catch exceptions from ``VWS-Python``."""
    assert not instance  # This is to satisfy the "vulture" linter.
    error_message = ""

    try:
        wrapped(*args, **kwargs)
    except UnknownTarget as exc:
        error_message = f'Error: Target "{exc.target_id}" does not exist.'
    except BadImage:
        error_message = (
            "Error: The given image is corrupted or the format is not "
            "supported."
        )
    except Fail:
        # We assume that exc.response.status_code == HTTPStatus.BAD_REQUEST
        error_message = (
            "Error: The request made to Vuforia was invalid and could not be "
            "processed. "
            "Check the given parameters."
        )
    except MetadataTooLarge:
        error_message = "Error: The given metadata is too large."
    except ImageTooLarge:
        error_message = "Error: The given image is too large."
    except TargetNameExist as exc:
        error_message = (
            f'Error: There is already a target named "{exc.target_name}".'
        )
    except ProjectInactive:
        error_message = (
            "Error: The project associated with the given keys is inactive."
        )
    except UnknownVWSErrorPossiblyBadName:
        error_message = (
            "Error: There was an unknown error from Vuforia. "
            "This may be because there is a problem with the given name."
        )
    except TargetStatusProcessing as exc:
        error_message = (
            f'Error: The target "{exc.target_id}" cannot be deleted as it is '
            "in the processing state."
        )
    except TargetStatusNotSuccess as exc:
        error_message = (
            f'Error: The target "{exc.target_id}" cannot be updated as it is '
            "in the processing state."
        )
    except AuthenticationFailure:
        error_message = "The given secret key was incorrect."
    except RequestTimeTooSkewed:
        error_message = (
            "Error: Vuforia reported that the time given with this request "
            "was outside the expected range. "
            "This may be because the system clock is out of sync."
        )
    # This exception is not available from the mock.
    except RequestQuotaReached:  # pragma: no cover
        error_message = (
            "Error: The maximum number of API calls for this database has "
            "been reached."
        )
    # This exception is not available from the mock.
    except DateRangeError:  # pragma: no cover
        error_message = (
            "Error: There was a problem with the date details given in the "
            "request."
        )
    # This exception is not available from the mock.
    except TargetQuotaReached:  # pragma: no cover
        error_message = (
            "Error: The maximum number of targets for this database has been "
            "reached."
        )
    # This exception is not available from the mock.
    except ProjectSuspended:  # pragma: no cover
        error_message = (
            "Error: The request could not be completed because this database "
            "has been suspended."
        )
    # This exception is not available from the mock.
    except ProjectHasNoAPIAccess:  # pragma: no cover
        error_message = (
            "Error: The request could not be completed because this database "
            "is not allowed to make API requests."
        )
    else:
        return

    click.echo(error_message, err=True)
    sys.exit(1)


def base_vws_url_option(command: Callable[..., None]) -> Callable[..., None]:
    """An option decorator for choosing the base VWS URL."""
    click_option_function: Callable[
        [Callable[..., None]],
        Callable[..., None],
    ] = click.option(
        "--base-vws-url",
        type=click.STRING,
        default="https://vws.vuforia.com",
        help="The base URL for the VWS API.",
        show_default=True,
    )

    function: Callable[..., None] = click_option_function(command)
    return function


@click.command(name="get-target-record")
@server_access_key_option
@server_secret_key_option
@target_id_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def get_target_record(
    server_access_key: str,
    server_secret_key: str,
    target_id: str,
    base_vws_url: str,
) -> None:
    """Get a target record.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API.html#How-To-Retrieve-a-Target-Record.
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )
    record = vws_client.get_target_record(target_id=target_id).target_record

    yaml_record = yaml.dump(dataclasses.asdict(record))
    click.echo(yaml_record)


@click.command(name="list-targets")
@server_access_key_option
@server_secret_key_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def list_targets(
    server_access_key: str,
    server_secret_key: str,
    base_vws_url: str,
) -> None:
    """List targets.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API.html#How-To-Get-a-Target-List-for-a-Cloud-Database.
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )
    targets = vws_client.list_targets()
    yaml_list = yaml.dump(targets)
    click.echo(yaml_list)


@click.command(name="get-duplicate-targets")
@server_access_key_option
@server_secret_key_option
@target_id_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def get_duplicate_targets(
    server_access_key: str,
    server_secret_key: str,
    target_id: str,
    base_vws_url: str,
) -> None:
    """Get a list of potential duplicate targets.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API.html#how-to-check-for-similar-targets.
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )
    record = vws_client.get_duplicate_targets(target_id=target_id)

    yaml_record = yaml.dump(record)
    click.echo(yaml_record)


@click.command(name="get-database-summary-report")
@server_access_key_option
@server_secret_key_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def get_database_summary_report(
    server_access_key: str,
    server_secret_key: str,
    base_vws_url: str,
) -> None:
    """Get a database summary report.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API.html#How-To-Get-a-Database-Summary-Report.
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )
    report = vws_client.get_database_summary_report()
    yaml_report = yaml.dump(dataclasses.asdict(report))
    click.echo(yaml_report)


@click.command(name="get-target-summary-report")
@server_access_key_option
@server_secret_key_option
@target_id_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def get_target_summary_report(
    server_access_key: str,
    server_secret_key: str,
    target_id: str,
    base_vws_url: str,
) -> None:
    """Get a target summary report.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API.html#How-To-Retrieve-a-Target-Summary-Report.
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )
    report = vws_client.get_target_summary_report(target_id=target_id)
    report_dict = dataclasses.asdict(report)
    report_dict["status"] = report_dict["status"].value
    report_dict["upload_date"] = str(report_dict["upload_date"])
    yaml_summary_report = yaml.dump(report_dict)
    click.echo(yaml_summary_report)


@click.command(name="delete-target")
@server_access_key_option
@server_secret_key_option
@target_id_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def delete_target(
    server_access_key: str,
    server_secret_key: str,
    target_id: str,
    base_vws_url: str,
) -> None:
    """Delete a target.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API.html#How-To-Delete-a-Target.
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )

    vws_client.delete_target(target_id=target_id)


@click.command(name="add-target")
@server_access_key_option
@server_secret_key_option
@target_name_option(command=None, required=True)
@target_width_option(command=None, required=True)
@target_image_option(command=None, required=True)
@application_metadata_option
@active_flag_option(command=None, allow_none=False)
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def add_target(
    server_access_key: str,
    server_secret_key: str,
    name: str,
    width: float,
    image_file_path: Path,
    active_flag_choice: ActiveFlagChoice,
    base_vws_url: str,
    application_metadata: str | None = None,
) -> None:
    """Add a target.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API#How-To-Add-a-Target
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )

    image_bytes = image_file_path.read_bytes()
    image = io.BytesIO(image_bytes)

    active_flag = {
        ActiveFlagChoice.TRUE: True,
        ActiveFlagChoice.FALSE: False,
    }[active_flag_choice]

    target_id = vws_client.add_target(
        name=name,
        width=width,
        image=image,
        active_flag=active_flag,
        application_metadata=application_metadata,
    )

    click.echo(target_id)


@click.command(name="update-target")
@server_access_key_option
@server_secret_key_option
@target_name_option(command=None, required=False)
@target_image_option(command=None, required=False)
@target_width_option(command=None, required=False)
@application_metadata_option
@active_flag_option(command=None, allow_none=True)
@target_id_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def update_target(
    server_access_key: str,
    server_secret_key: str,
    target_id: str,
    image_file_path: Path | None,
    base_vws_url: str,
    name: str | None = None,
    application_metadata: str | None = None,
    active_flag_choice: ActiveFlagChoice | None = None,
    width: float | None = None,
) -> None:
    """Update a target.

    \b
    See
    https://library.vuforia.com/articles/Solution/How-To-Use-the-Vuforia-Web-Services-API#How-To-Update-a-Target
    """
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )

    if image_file_path is None:
        image = None
    else:
        image_bytes = image_file_path.read_bytes()
        image = io.BytesIO(image_bytes)

    active_flag = {
        ActiveFlagChoice.TRUE: True,
        ActiveFlagChoice.FALSE: False,
        None: None,
    }[active_flag_choice]

    vws_client.update_target(
        name=name,
        target_id=target_id,
        image=image,
        application_metadata=application_metadata,
        active_flag=active_flag,
        width=width,
    )


_SECONDS_BETWEEN_REQUESTS_DEFAULT = 0.2

_SECONDS_BETWEEN_REQUESTS_HELP = (
    "The number of seconds to wait between requests made while polling the "
    "target status. "
    f"We wait {_SECONDS_BETWEEN_REQUESTS_DEFAULT} seconds by default, rather "
    "than less, than that to decrease the number of calls made to the API, to "
    "decrease the likelihood of hitting the request quota."
)

_TIMEOUT_SECONDS_HELP = (
    "The maximum number of seconds to wait for the target to be processed."
)


@click.command(name="wait-for-target-processed")
@click.option(
    "--seconds-between-requests",
    type=click.FloatRange(min=0.05),
    default=_SECONDS_BETWEEN_REQUESTS_DEFAULT,
    help=_SECONDS_BETWEEN_REQUESTS_HELP,
    show_default=True,
)
@click.option(
    "--timeout-seconds",
    type=click.FloatRange(min=0.05),
    default=300,
    help=_TIMEOUT_SECONDS_HELP,
    show_default=True,
)
@server_access_key_option
@server_secret_key_option
@target_id_option
@handle_vws_exceptions  # type: ignore[misc]
@base_vws_url_option
def wait_for_target_processed(
    server_access_key: str,
    server_secret_key: str,
    target_id: str,
    seconds_between_requests: float,
    base_vws_url: str,
    timeout_seconds: float,
) -> None:
    """Wait for a target to be "processed". This is done by polling the VWS API."""
    vws_client = VWS(
        server_access_key=server_access_key,
        server_secret_key=server_secret_key,
        base_vws_url=base_vws_url,
    )

    try:
        vws_client.wait_for_target_processed(
            target_id=target_id,
            seconds_between_requests=seconds_between_requests,
            timeout_seconds=timeout_seconds,
        )
    except TargetProcessingTimeout:
        click.echo(f"Timeout of {timeout_seconds} seconds reached.", err=True)
        sys.exit(1)



"""Create binaries for the CLIs."""

import logging
import uuid
from pathlib import Path

import docker
from docker.types import Mount

LOGGER = logging.getLogger(__name__)


def make_linux_binaries(repo_root: Path) -> None:
    """Create binaries for Linux in a Docker container.

    Args:
        repo_root: The path to the root of the repository.
    """
    client = docker.from_env()
    dist_dir = repo_root / "dist"
    assert not dist_dir.exists() or not set(dist_dir.iterdir())

    target_dir = "/" + uuid.uuid4().hex
    code_mount = Mount(
        source=str(repo_root.absolute()),
        target=target_dir,
        type="bind",
    )

    # We install in editable mode to overwrite any potential
    # ``_setuptools_scm_version.txt`` file.
    cmd_in_container = " ".join(
        [
            "pip",
            "install",
            "--editable",
            ".[packaging]",
            "&&",
            "python",
            "admin/create_pyinstaller_binaries.py",
        ],
    )
    command = f'bash -c "{cmd_in_container}"'

    container = client.containers.run(
        image="python:3.11",
        mounts=[code_mount],
        command=command,
        working_dir=target_dir,
        remove=True,
        detach=True,
    )
    for line in container.logs(stream=True):
        line = line.decode().strip()
        LOGGER.warning(line)

    status_code = container.wait()["StatusCode"]
    assert status_code == 0
